/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : alarm.c                                                  */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Provides timer functions for use by timer node.          */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <timer.h> /* Specific header for this dll                            */

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <threads.h>
#include <time.h>
#include <cmqc.h>

void genGMTDateAndTime(OUT char *dateString,
                       OUT char *timeString);

/*
 * Reset the alarm element for a new timer 
 */
void setTimerTimeoutAlarm(IN timerInstance_t *timerInstanceP)
{

	/*
	 * Assumption:
	 *
	 * On entry to this function, the caller has locked the timer chain
	 */
	time_t currentTime;
	int timeoutInterval;
	timerInstance_t *currentTimerP;

    TRACE_ENTRY("setTimerTimeoutAlarm", ("timerInstanceP=%.8x", timerInstanceP));

	/* calculate the new timeout interval */
	timeoutInterval = timerInstanceP->intervalSeconds;
    TRACE_MID(TRACE_INFO, ("Timeout interval is %d", timeoutInterval));
	time(&currentTime);
	timerInstanceP->timeoutTime = currentTime + timeoutInterval;
    TRACE_MID(TRACE_INFO, ("Timeout value is %d", timerInstanceP->timeoutTime));

	/* find position in alarm chain for this timer based on its expiry time */
    currentTimerP = &instanceAnchor;
    TRACE_MID(TRACE_INFO, ("Starting search of alarm chain at anchor (%.8x)",
		                   currentTimerP));

	while(TRUE)
	{
        TRACE_MID(TRACE_INFO, ("Checking timer instance at %.8x "
			                   "(timeout value = %d)", currentTimerP,
							   currentTimerP->timeoutTime));

		if (!currentTimerP->nextAlarmP ||
			currentTimerP->nextAlarmP->timeoutTime > timerInstanceP->timeoutTime)
		{
            TRACE_MID(TRACE_INFO, ("Inserting alarm after timer at %.8x",
				                   currentTimerP));
			break;
		}

		currentTimerP = currentTimerP->nextAlarmP;
	}

    /* add after current alarm member */
	INSERT_AFTER_IN_ALARM_CHAIN(currentTimerP, timerInstanceP);
    TRACE_RETURN_VOID(RSN_001);
}

typedef struct handleEntry {
	char *queueName;
	MQHOBJ queueHandle;
} handleEntry_t;

handleEntry_t *queueHandleTable = NULL;
int numQueueHandleSlots = 0,
	numQueueHandleEntries = 0;

/******************************************************************************/
/* processAlarms                                                              */
/******************************************************************************/
threadFunction_t processAlarms(void *dataP)
{
    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
	MQHCONN qmConn = 0;
	char errorMessage[1000];
	time_t currentTime;
	int timeToWait,
		retCode;
	bool_t timerPopThresholdReached;
	timerInstance_t *thisAlarmP,
                    *nextAlarmP;
	globalTimerInstance_t *globalAlarmP;

    TRACE_ENTRY("processAlarms", ("dataP=%.8x", dataP));

	/* get handle to queue manager */
    TRACE_MID(TRACE_INFO, ("About to issue MQCONN for '%s'",
		                    queueManagerName));
	MQCONN(queueManagerName,
		   &qmConn,                 /* connection handle */
		   &compCode,               /* completion code */
		   &reason);                /* reason code */

	if (compCode == MQCC_FAILED)
	{
		/* create error message to pass back to parent */
		sprintf(errorMessage, "MQCONN for queue manager '%s' "
                           "failed with reason code %ld",
                           queueManagerName,
                           reason);

        TRACE_MID(TRACE_INFO, ("%s", errorMessage));
        alarmThreadErrorMessage = strdup(errorMessage);

		/* lock the alarm thread sync mutex and post the parent thread */
	    threadMutexLock(alarmThreadSyncMutex);
		threadCondSignal(alarmThreadSyncCondVar);
	    threadMutexUnlock(alarmThreadSyncMutex);

        TRACE_RETURN_TYPE(RSN_002, FUNCTION_RETURN_TYPE, 0);
	}

	/* let the parent thread know that we're up and running */
	alarmThreadInitialised = TRUE;

	/* lock the alarm thread sync mutex and post the parent thread */
    TRACE_MID(TRACE_INFO, ("About to signal parent for initialisation complete"));
	threadMutexLock(alarmThreadSyncMutex);
	threadCondSignal(alarmThreadSyncCondVar);
	threadMutexUnlock(alarmThreadSyncMutex);

	/* 
	 * The timer chain and the alarm chain are closely related, so we lock
	 * both via a single mutex.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
	threadMutexLock(timerChainMutex);

	/* keep processing alarm chain until requested to shutdown */
	while(!shutdownAlarmRequested)
	{
        TRACE_MID(TRACE_INFO, ("Processing alarm element chain"));

		thisAlarmP = instanceAnchor.nextAlarmP;

		/* process any expired alarm elements */
	    while(thisAlarmP)
		{
 		    /* run down the list of alarm elements */
            TRACE_MID(TRACE_INFO, ("Processing alarm element at %.8x",
				                    thisAlarmP));

			/* ignore any timers that have expired */
			if (thisAlarmP->expired)
			{
	            TRACE_MID(TRACE_INFO, ("Ignoring expired timer at %.8x", thisAlarmP));
				thisAlarmP = thisAlarmP->nextAlarmP;
				continue;
			}

			/* address next alarm element, just in case we need to free the current */
            TRACE_MID(TRACE_INFO, ("thisAlarmP->nextAlarmP = %.8x",
				                    thisAlarmP->nextAlarmP));
			nextAlarmP = thisAlarmP->nextAlarmP;

			/* 
			 * if the timer interval has expired, then generate a timer event message
			 */
			time(&currentTime);
            TRACE_MID(TRACE_INFO, ("thisAlarmP->timeoutTime = %d, currentTime=%d",
				                    thisAlarmP->timeoutTime, currentTime));

			timerPopThresholdReached = FALSE;
			if (thisAlarmP->timeoutTime <= currentTime)
			{
				/* decrement the timerPopCount if necessary */
				if (thisAlarmP->timerPopCount != INFINITE)
				{
					if (thisAlarmP->global)
					{
						/* take the global lock */
						shmLock(thisAlarmP->scope==SCOPE_BROKER);

						globalAlarmP = (globalTimerInstance_t *)thisAlarmP;
						globalAlarmP->timerPopCount--;

						/* have we reached the timer pop threshold? */
						if (!globalAlarmP->timerPopCount)
						{
				            TRACE_MID(TRACE_INFO, ("Global timer pop count threshold reached"));
							timerPopThresholdReached = TRUE;
							globalAlarmP->expired = TRUE;
						}

						/* release the global lock */
						shmUnlock(thisAlarmP->scope==SCOPE_BROKER);
					}
					else
					{
						thisAlarmP->timerPopCount--;
						if (!thisAlarmP->timerPopCount)
						{
				            TRACE_MID(TRACE_INFO, ("Timer pop count threshold reached"));
							timerPopThresholdReached = TRUE;
							thisAlarmP->expired = TRUE;
						}
					}
				}

				/* remove the alarm element from the alarm chain */
				REMOVE_FROM_ALARM_CHAIN(thisAlarmP);

				/* generate a timer message */
				retCode = generateTimerEventMessage(qmConn,
					                                thisAlarmP);
				if (retCode != TT_SUCCESS)
				{
				    /* not sure what to do if this function fails... */
				}

				/* 
				 * remove timer if its expired
				 */
				if (timerPopThresholdReached)
				{
		            TRACE_MID(TRACE_INFO, ("Removing expired timer"));
					if (thisAlarmP->global)
					{
						shmLock(thisAlarmP->scope==SCOPE_BROKER);
					}
					freeTimer(thisAlarmP);
					if (thisAlarmP->global)
					{
						shmUnlock(thisAlarmP->scope==SCOPE_BROKER);
					}
				}
				else
				{
					/* reset the timeout value for the next timer interval */
					setTimerTimeoutAlarm(thisAlarmP);
				}

				thisAlarmP = nextAlarmP;
				continue;
			}

			/* alarms are in order of expiry time...so nothing more to do here */
			break;
		}

		/* reset the alarm thread to pop at next expiry interval */
		if (instanceAnchor.nextAlarmP)
		{
			/* reset the alarm for the next element (milliseconds) */
			timeToWait = 
				(instanceAnchor.nextAlarmP->timeoutTime - currentTime) * 1000;
            TRACE_MID(TRACE_INFO, ("Next expiry interval time is %d", timeToWait));
		}
		else
		{
			timeToWait = INFINITE;
            TRACE_MID(TRACE_INFO, ("Next expiry interval time is INFINITE"));
		}

	    /* wait for next alarm to pop or a new element to appear on the chain */
        TRACE_MID(TRACE_INFO, ("Alarm thread about to sleep for expiry interval"));
		threadCondTimedWait(alarmThreadWorkCondVar, timeToWait);
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, (">>>>>>> Alarm thread woken <<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<"));
	}

	if (qmConn)
	{
		if (numQueueHandleSlots)
		{
			TRACE_MID(TRACE_INFO, ("Freeing up queue handles"));

			/* see if we've already got a handle to this queue */
			numQueueHandleEntries = 0;  /* remove all slots */
			numQueueHandleSlots = 0;  /* remove all slots */
			free(queueHandleTable);
			queueHandleTable = NULL;
		}

		TRACE_MID(TRACE_INFO, ("About to issue MQDISC for '%s'",
								queueManagerName));
		MQDISC(&qmConn,                 /* connection handle */
			   &compCode,               /* completion code */
			   &reason);                /* reason code */
	}

    TRACE_MID(TRACE_INFO, ("Alarm thread terminating"));

	threadMutexUnlock(timerChainMutex);

    TRACE_RETURN_TYPE(RSN_001, FUNCTION_RETURN_TYPE, 0);
}


/******************************************************************************/
/* generateTimerEventMessage                                                  */
/******************************************************************************/
int generateTimerEventMessage(IN MQHCONN qmConn,
						      IN timerInstance_t *timerInstanceP)
{
    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
    MQLONG   openOptions;               /* MQOPEN options                */
	MQHOBJ queueHandle;
    MQOD od = {MQOD_DEFAULT};           /* Object Descriptor */
    MQMD md = {MQMD_DEFAULT};           /* Message Descriptor */
    MQPMO pmo = {MQPMO_DEFAULT};        /* put message options */
	CciChar *UC_queueNameP;
	char *storedDataP,
		 *payloadP,
		 *endPos,
		 *currentPos,
		 queueNameMB[MAX_STRING_SIZE+1],
		 *eventQueueP,
	     *XMLCurrentPos,
		 dateString[4+2+2+1],
		 timeString[2+2+2+2+1];
	int i,
		retCode,
		payloadLength,
		depth = 0;
	bool_t treeStructuredData;
	globalTimerInstance_t *globalTimerInstanceP;

    TRACE_ENTRY("generateTimerEventMessage", ("qmConn=%.8x, timerInstanceP=%.8x",
		                                      qmConn,
									    	  timerInstanceP));

	/* build a timer event message and PUT to the specified queue */

	/* address the queue name and convert to multibyte string */

	/* send timed out message to the reply node input queue */
	if (timerInstanceP->global)
	{
		globalTimerInstanceP = (globalTimerInstance_t *)timerInstanceP;

		UC_queueNameP = SHM_HANDLE2ADDRESS(globalTimerInstanceP->queueNameP,
			                               CciChar *, 
										   globalTimerInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
        TRACE_MID(TRACE_INFO, ("Global address for queue name at %.8x", UC_queueNameP));
	}
	else
	{
		UC_queueNameP = timerInstanceP->deployedTimerP->queueName;
	}
	UCsprintf(queueNameMB, "%S", UC_queueNameP);
	eventQueueP = queueNameMB;

	/* see if we've already got a handle to this queue */
	queueHandle = 0;
    TRACE_MID(TRACE_INFO, ("Searching for stored queue handle for '%s'",
		                    eventQueueP));
	for (i=0; i<numQueueHandleEntries; i++)
	{
        TRACE_MID(TRACE_INFO, ("Processing stored queue handle for '%s'",
			                    (queueHandleTable+i)->queueName));
		if (!strcmp((queueHandleTable+i)->queueName,
			         eventQueueP))
		{
            TRACE_MID(TRACE_INFO, ("Found stored queue handle"));
			queueHandle = (queueHandleTable+i)->queueHandle;
		}
	}

	if (!queueHandle)
	{
		/* open the queue and store the handle */

		/* initialise the object descriptor with the queue name */
		strncpy(od.ObjectName, 
                eventQueueP,
				(size_t)MQ_Q_NAME_LENGTH);

		/* set open mode for writing messages */
		openOptions = MQOO_OUTPUT;
		openOptions += MQOO_SET_ALL_CONTEXT;       /* we need to set PutApplName */
		openOptions += MQOO_FAIL_IF_QUIESCING;     /* but not if MQM stopping */
		 
		TRACE_MID(TRACE_INFO, ("About to issue MQOPEN for '%s'",
								eventQueueP));

		MQOPEN(qmConn,              /* connection handle */
			   &od,                 /* object descriptor for queue */
			   openOptions,         /* open options */
			   &queueHandle,        /* object handle */
			   &compCode,           /* completion code */
			   &reason);            /* reason code */
            
		if (compCode == MQCC_FAILED)
		{
			TRACE_MID(TRACE_ERROR, ("MQOPEN for '%s' failed with "
								  "reason code %ld",
								  eventQueueP,
								  reason));
            TRACE_RETURN(RSN_004, TT_FAILED);
		}

		/* increase handle table if full */
		if (numQueueHandleSlots == numQueueHandleEntries)
		{
			numQueueHandleSlots +=10;  /* increase by 10 slots */
	        TRACE_MID(TRACE_INFO, ("Increasing handle table to %d slots",
							       numQueueHandleSlots));
            queueHandleTable = (handleEntry_t *)realloc(queueHandleTable,
							   (numQueueHandleSlots)*(sizeof(handleEntry_t)));
		}

		/* add queue handle to table */
        TRACE_MID(TRACE_INFO, ("Queue handle for '%s' added to slot %d",
						       eventQueueP,
							   numQueueHandleEntries));
		(queueHandleTable+numQueueHandleEntries)->queueName = strdup(eventQueueP);
		(queueHandleTable+numQueueHandleEntries)->queueHandle = queueHandle;
        numQueueHandleEntries++;
	}

	/* create message data */

	/* allocate an initial buffer for the XML message */
	payloadLength = ONE_KBYTE;
	payloadP = (char *)malloc(payloadLength);
    TRACE_MID(TRACE_INFO, ("Initial message buffer allocated at %.8x", payloadP));
	XMLCurrentPos = payloadP;

	if (timerInstanceP->savedDataLength)
	{
		/* create enclosing XML tags */
		strcpy(payloadP, TIMER_DATA_BEGIN_TAG);
		XMLCurrentPos += strlen(TIMER_DATA_BEGIN_TAG);

		treeStructuredData = FALSE;
		/* address the data buffer */
		if (timerInstanceP->global)
		{
			globalTimerInstanceP = (globalTimerInstance_t *)timerInstanceP;
			storedDataP = SHM_HANDLE2ADDRESS(globalTimerInstanceP->savedDataP, char *, globalTimerInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			if (globalTimerInstanceP->retrieveMessageDataFromTree)
			{
				treeStructuredData = TRUE;
			}
		}
		else
		{
			storedDataP = timerInstanceP->savedDataP;
			if (timerInstanceP->deployedTimerP->retrieveMessageDataFromTree)
			{
				treeStructuredData = TRUE;
			}
		}

		/* calculate the end of the data buffer */
		endPos = storedDataP + timerInstanceP->savedDataLength;

		/* if data is in tree structure, then serialise it into buffer */
		if (treeStructuredData)
		{
		    TRACE_MID(TRACE_INFO, ("Message data is a syntax tree structure"));

			/* run through the serialised data and build it into an XML message */
			currentPos = storedDataP;
			retCode = convertSyntaxElementsToXML(&payloadP,
												 &payloadLength,
									 			 &XMLCurrentPos,
												 storedDataP,
												 timerInstanceP->savedDataLength,
												 &currentPos,
												 endPos,
												 &depth);
		}
		else
		{
		    TRACE_MID(TRACE_INFO, ("Message data is a literal"));
			/* copy the data into XML message */

			/* make sure there's enough space in the buffer for the data */
			extendBufferIfRequired(strlen(storedDataP),
								   &payloadP,
								   &XMLCurrentPos,
								   &payloadLength);

			memcpy(XMLCurrentPos, storedDataP, strlen(storedDataP));
			XMLCurrentPos += strlen(storedDataP);
		}

		/* make sure there's enough space in the buffer for the closing tag */
		extendBufferIfRequired(strlen(TIMER_DATA_END_TAG),
							   &payloadP,
							   &XMLCurrentPos,
							   &payloadLength);
			
		/* copy in the closing XML tag */
		memcpy(XMLCurrentPos, TIMER_DATA_END_TAG, strlen(TIMER_DATA_END_TAG));
		XMLCurrentPos += strlen(TIMER_DATA_END_TAG);

		/* calculate the total size of the XML message */
		TRACE_MID(TRACE_INFO, ("XMLCurrentPos is %.8x, payloadP is %.8x", 
				                XMLCurrentPos, payloadP));
		payloadLength = (unsigned long)XMLCurrentPos - (unsigned long)(payloadP);
		TRACE_MID(TRACE_INFO, ("Calculated XML message length is %d", payloadLength));
	}
	else
	{
		strcpy(payloadP, TIMER_DATA_NULL_TAG);
		payloadLength = strlen(TIMER_DATA_NULL_TAG);
	}

	/* send timer event message to specified queue */
    md.Persistence = MQPER_PERSISTENT;

	/*
	 * It would be desirable to be able to use the single SYSTEM.BROKER.TIMER.QUEUE for
	 */

	/* set correl id to message id of message which created the timer */
	memcpy(md.CorrelId, timerInstanceP->msgId, sizeof(md.CorrelId));

	/* set PutApplName value to 'AggregatorTimerThread' */
	memcpy(md.PutApplName, TIMER_THREAD_ID_MB, strlen(TIMER_THREAD_ID_MB));

	/* set expiry if specified */
	if (timerInstanceP->msgExpiry != INFINITE) {
		TRACE_MID(TRACE_INFO,  ("Setting timer message expiry to %d seconds",
			                     timerInstanceP->msgExpiry));
	    md.Expiry = timerInstanceP->msgExpiry * 10; // expressed as tenths of a second
	}

	/* make sure the PutApplName value is used on input */
	pmo.Options += MQPMO_SET_ALL_CONTEXT;

	/* set the PutDate, PutTime and PutApplType values */
	md.PutApplType = MQAT_DEFAULT;

    genGMTDateAndTime(dateString, timeString);
    TRACE_MID(TRACE_INFO, ("Date string: '%s'", dateString));
    TRACE_MID(TRACE_INFO, ("Time string: '%s'", timeString));
	memcpy(md.PutDate, dateString, sizeof(md.PutDate));
	memcpy(md.PutTime, timeString, sizeof(md.PutTime));
	
    TRACE_MID(TRACE_INFO, ("About to write timer event message"));
    
    /* put the message on the queue */
    MQPUT(qmConn,                   /* connection handle */
          queueHandle,              /* object handle */
          &md,                      /* message descriptor */
          &pmo,                     /* put options */
          payloadLength,            /* msg length */
          payloadP,                 /* message buffer */
          &compCode,                /* completion code */
          &reason);                 /* reason code */
        
	if (compCode != MQCC_OK)
	{
		TRACE_MID(TRACE_ERROR, ("MQPUT for '%s' failed with "
							  "reason code %ld",
							  eventQueueP,
							  reason));
        TRACE_RETURN(RSN_005, TT_FAILED);
	}
    TRACE_MID(TRACE_INFO, ("Timer event message written to queue"));
	if (payloadP)
	{
	    TRACE_MID(TRACE_INFO, ("Free'ing message buffer"));
		free(payloadP);
	}

    TRACE_RETURN(RSN_006, TT_SUCCESS);
}

/*********************/
/* genGMTDateAndTime */
/*********************/
#include <time.h>
void genGMTDateAndTime(OUT char *dateString,
                       OUT char *timeString)
{
    time_t thisTime;
    struct tm *newTime;

	TRACE_ENTRY("genGMTDateAndTime", (" "));

    /* get the date and time info */
    time(&thisTime);
    newTime = gmtime(&thisTime);

    /* 
     * The gmtime function produces a structure containing values for 
     * day, month, year, etc. as UTC
     *
     * NB: will return null pointer for dates prior to midnight 1 Jan 1970
     */

    if (newTime)
    {
        /* extract the bits we need */

        /* year is value less 1900, so we need to qualify it */
        if (newTime->tm_year < 90)
        {
            newTime->tm_year += 2000;
        }
        else
        {
            newTime->tm_year += 1900;
        }

        /* month is 0 - 11, so we bump it by one */
        newTime->tm_mon++;

        sprintf(dateString, "%04.4d%02.2d%02.2d",
                            newTime->tm_year,
                            newTime->tm_mon,
                            newTime->tm_mday);
        sprintf(timeString, "%02.2d%02.2d%02.2d00", /* tenths=0, hundredths=0 */
                            newTime->tm_hour,
                            newTime->tm_min,
                            newTime->tm_sec);
    }
    else
    {
        TRACE_MID(TRACE_ERROR, ("Error extracting time via localtime"));
        dateString = "";
        timeString = "";
    }

    TRACE_RETURN_VOID(RSN_007)
}


