/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : create.c                                                 */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Creates a timer instance, optionally storing data from   */
/*                   the current message for inclusion in any timer message(s)*/
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: July 2001                      */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <timer.h>      /* Specific header for this dll                      */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

/* local function prototypes */
bool_t isNumeric(char *string);

/******************************************************************************/
/* Node Utility Functions                                                     */
/*                                                                            */
/* insertAttributeTableEntry                                                  */
/* insertInputTerminalListEntry                                               */
/* insertOutputTerminalListEntry                                              */
/* getInputTerminalHandle                                                     */
/* getOutputTerminalHandle                                                    */
/* traverseMessage                                                                           */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* evaluateTimerCreate                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates a timer instance based on attributes of the current message. These */
/* include the MsgId and CorrelId and (optionally) additional data from the   */
/* message.                                                                   */
/*                                                                            */
/******************************************************************************/
int evaluateTimerCreate(context_t* contextP,
 			   		    CciMessage *destinationList,
		                CciMessage *exceptionList,
		                CciMessage *message)
{
	void *correlIdElementP,
		 *messageIdElementP;
	timerInstance_t *timerInstanceP,
		                   *instanceAnchorP;
	globalTimerInstance_t *globalTimerInstanceP;
    struct CciByteArray byteArrayDetails;
    CciSize numBytes;
	int rc,
		retCode,
		savedDataBufferLength,
		valueBufferLength,
		allocSize,
		derivedIntervalSeconds,
		messageNumber;
	deployedTimer_t *deployedTimerP;
	CciElement *copyFromElement;
	char *savedDataBuffer,
		 *valueBuffer,
	     *timerNameP,
		 *dataValueP,
		 mbString[MAX_STRING_SIZE + 1],
		 msgBuffer[1000];
	bool_t derivedIntervalValue = FALSE,
		   issueLogMessage = FALSE;
	scope_t scope;
    char *executionGroupNameP;
	CciChar *messageFlowNameP,
			*queueNameP;
	void *dataP;
	char nullId[MQ_MSG_ID_LENGTH];
    static char *functionName="evaluateTimerCreate";

    TRACE_ENTRY("evaluateTimerCreate", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	deployedTimerP = (deployedTimer_t *)(contextP->nodeDataP);

	scope = deployedTimerP->scope;
	TRACE_MID(TRACE_INFO, ("Timer scope is %s",
		                    scope==SCOPE_EXECUTION_GROUP ? "LOCAL" : "GLOBAL"));

	/* Create a new local timer instance.  Even if its a global timer, we build a local
	 * copy first since this makes pointer manipulation somewhat easier.  For global timers
	 * we allocate global storage later and then transfer data from the local copy.
	 */
	timerInstanceP = (timerInstance_t *)malloc(sizeof(timerInstance_t));

	memset(timerInstanceP, 0, sizeof(timerInstance_t));
	TRACE_MID(TRACE_INFO, ("New %s timer instance created at %.8x", 
 		                   scope==SCOPE_EXECUTION_GROUP ? "LOCAL" : "GLOBAL",
		                   timerInstanceP));

	/* extract the message id and correl id values */
	messageIdElementP = getMessageElement(message, MESSAGE_ID);

	if (!messageIdElementP)
	{
		free(timerInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Message Id not found in message"));
		TRACE_RETURN(RSN_014, TT_MESSAGE_ID_NOT_FOUND);
	}

	/* copy message id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)(timerInstanceP->msgId);
	byteArrayDetails.size = sizeof(timerInstanceP->msgId);
	numBytes = cniElementByteArrayValue(&rc,
										messageIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		free(timerInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Message Id from message"));
		TRACE_RETURN(RSN_015, TT_MESSAGE_ID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Message Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, timerInstanceP->msgId, MQ_MSG_ID_LENGTH);

	/* extract the correlId */
	correlIdElementP = getMessageElement(message, CORREL_ID);

	if (!correlIdElementP)
	{
		free(timerInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Correlation Id not found in message"));
		TRACE_RETURN(RSN_014, TT_CORRELID_NOT_FOUND);
	}

	/* copy correl id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)(timerInstanceP->correlId);
	byteArrayDetails.size = sizeof(timerInstanceP->correlId);
	numBytes = cniElementByteArrayValue(&rc,
										correlIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		free(timerInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Correl Id from message"));
		TRACE_RETURN(RSN_015, TT_CORRELID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Correl Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, timerInstanceP->correlId, MQ_CORREL_ID_LENGTH);

	/* initialise other timer attributes */
	timerInstanceP->deployedTimerP = contextP->nodeDataP;
	timerInstanceP->scope = scope;
	timerInstanceP->timerPopCount = timerInstanceP->deployedTimerP->maxTimerPopCount;
	timerInstanceP->msgExpiry = timerInstanceP->deployedTimerP->msgExpiry;

	/* process TimerName if derived from syntax tree */
	valueBufferLength = 0;
	valueBuffer = NULL;
	if (timerInstanceP->deployedTimerP->retrieveTimerNameFromTree)
	{
		retCode = getValueFromSyntaxTree(message,
										 timerInstanceP->deployedTimerP->timerName,
										 TRUE, // allow child VALUE type node to be used
										 &valueBuffer,
										 &valueBufferLength);

		if (retCode != TT_SUCCESS)
		{
			/* deal with common errors */
			if (retCode == TT_ELEMENT_NOT_FOUND)
			{
				sprintf(msgBuffer, "Failed to locate syntax element specified for the '%S' "
								   "attribute in message.  The element path is '%S'",
								   ATTR_TIMER_NAME_MB,
								   timerInstanceP->deployedTimerP->timerName);
				messageNumber = MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT;
			}
			else
			if (retCode == TT_ELEMENT_HAS_NO_VALUE)
			{
				sprintf(msgBuffer, "Failed to locate any data in syntax element specified "
					               "for the '%S' attribute in message.  The element path is '%S'",
								   ATTR_TIMER_NAME_MB,
								   timerInstanceP->deployedTimerP->timerName);
				messageNumber = MJB_SYNTAX_ELEMENT_CONTAINS_NO_DATA;
			}
			else
			if (retCode == TT_ELEMENT_DATA_CONTAINS_NULL_DATA_CHARACTER)
			{
				sprintf(msgBuffer, "Failed to extract any data from syntax element specified "
					               "for the '%S' attribute in message because the data "
								   "contains one or more null characters.  The element path "
								   "is '%S'",
								   ATTR_TIMER_NAME_MB,
								   timerInstanceP->deployedTimerP->timerName);
				messageNumber = MJB_SYNTAX_ELEMENT_DATA_CONTAINS_NULLS;
			}
			else
			{
				/* some other error has occurred */
				sprintf(msgBuffer, "Error occurred when trying to extract data from syntax "
					               "element specified for the '%S' attribute in message.  "
								   "The element path is '%S'",
								   ATTR_TIMER_NAME_MB,
								   timerInstanceP->deployedTimerP->timerName);
				messageNumber = MJB_SYNTAX_ELEMENT_DATA_RETRIEVAL_FAILURE;
			}

			/* issue the appropriate log message */
			cciLog(&rc, 
				   CCI_LOG_ERROR, 
				   __FILE__, 
				   __LINE__, 
				   functionName, 
				   C.PLUGIN_MESSAGE_CATALOGUE, 
				   messageNumber, 
				   msgBuffer,
#if !defined(WIN32)
				   brokerName,
#endif
				   ATTR_TIMER_NAME_MB,
 				   /* convert MessageData to multi byte string */
				   UCToStr(timerInstanceP->deployedTimerP->timerName,
						   mbString,
						   sizeof(mbString),
						   NULL),
				   (char *)0);
			                     

			free(timerInstanceP);

			TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
								   "terminal", message));
			/* 
			 * Propagate message to the 'failure' terminal if element doesn't exist or 
			 * has no data.
			 */
			retCode = propagateMessage(contextP,
									   C.TERM_FAILURE,
									   message,
									   destinationList,
									   exceptionList,
									   FALSE); /* this is not a new message */

			if (retCode != TT_SUCCESS)
			{
				TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
				RETHROW_IF_EXCEPTION(retCode);
				TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
			}

			TRACE_RETURN(RSN_013, TT_FAILED);
		}

		/* store timer name in timer instance */
        timerInstanceP->timerName = valueBuffer;
		TRACE_MID(TRACE_INFO, ("Derived timer name '%s' stored in timer instance",
			                   valueBuffer)); 
	}
	else
	if (timerInstanceP->deployedTimerP->timerName)
	{
		valueBuffer = (char *)malloc(UClen(timerInstanceP->deployedTimerP->timerName) +1);
        UCToStr(timerInstanceP->deployedTimerP->timerName,
                valueBuffer,
                UClen(timerInstanceP->deployedTimerP->timerName) +1,
				NULL);

		timerInstanceP->timerName = valueBuffer;
		TRACE_MID(TRACE_INFO, ("Literal timer name '%s' stored in timer instance", valueBuffer)); 
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Timer instance has no name"));

		/* reject any timer with no identifiers, e.g. no name, null MsgId and null CorrelId */
		memset(nullId, 0, sizeof(nullId));
		if (!memcmp(timerInstanceP->msgId, nullId, sizeof(nullId)) &&
		    !memcmp(timerInstanceP->correlId, nullId, sizeof(nullId)))
		{
			TRACE_MID(TRACE_ERROR, ("Invalid timer has no identifiers (no name, null Msgid and null CorrelId)"));

			// log error message and propagate message to failure terminal
			sprintf(msgBuffer, "Attempt to create a timer with no identifiers has failed.  Timer "
				               "has no name, MsgId or CorrelId values");
			TRACE_MID(TRACE_ERROR, ("%s", msgBuffer)); 

			/* issue the appropriate log message */
			cciLog(&rc, 
				   CCI_LOG_ERROR, 
				   __FILE__, 
				   __LINE__, 
				   functionName, 
				   C.PLUGIN_MESSAGE_CATALOGUE, 
				   MJB_CREATE_WITH_NO_IDENTIFIER, 
				   msgBuffer,
#if !defined(WIN32)
				   brokerName,
#endif
				   (char *)0);

			free(timerInstanceP);

			TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
								   "terminal", message));
			/* 
			 * Propagate message to the 'failure' terminal if element doesn't exist or 
			 * has no data.
			 */
			retCode = propagateMessage(contextP,
									   C.TERM_FAILURE,
									   message,
									   destinationList,
									   exceptionList,
									   FALSE); /* this is not a new message */

			if (retCode != TT_SUCCESS)
			{
				TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
				RETHROW_IF_EXCEPTION(retCode);
				TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
			}

			TRACE_RETURN(RSN_013, TT_FAILED);
		}
	}

	/* extract IntervalSeconds if derived from syntax tree */
	valueBufferLength = 0;
	valueBuffer = NULL;
	dataValueP = NULL;
	if (timerInstanceP->deployedTimerP->retrieveIntervalSecondsFromTree)
	{
		TRACE_MID(TRACE_INFO, ("Extracting %s value from syntax tree", ATTR_INTERVAL_SECONDS_MB));
		retCode = getValueFromSyntaxTree(message,
										 timerInstanceP->deployedTimerP->intervalSecondsElement,
										 TRUE, // allow child VALUE type node to be used
										 &valueBuffer,
										 &valueBufferLength);

		if (retCode != TT_SUCCESS)
		{
			issueLogMessage = TRUE;
			/* deal with common errors */
			if (retCode == TT_ELEMENT_NOT_FOUND)
			{
				sprintf(msgBuffer, "Failed to locate syntax element specified for the '%S' "
								   "attribute in message.  The element path is '%S'",
								   ATTR_INTERVAL_SECONDS_MB,
								   timerInstanceP->deployedTimerP->intervalSecondsElement);
				messageNumber = MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT;
			}
			else
			if (retCode == TT_ELEMENT_HAS_NO_VALUE)
			{
				sprintf(msgBuffer, "Failed to locate any data in syntax element specified "
					               "for the '%S' attribute in message.  The element path is '%S'",
								   ATTR_INTERVAL_SECONDS_MB,
								   timerInstanceP->deployedTimerP->intervalSecondsElement);
				messageNumber = MJB_SYNTAX_ELEMENT_CONTAINS_NO_DATA;
			}
			else
			if (retCode == TT_ELEMENT_DATA_CONTAINS_NULL_DATA_CHARACTER)
			{
				sprintf(msgBuffer, "Failed to extract any data from syntax element specified "
					               "for the '%S' attribute in message because the data "
								   "contains one or more null characters.  The element path "
								   "is '%S'",
								   ATTR_INTERVAL_SECONDS_MB,
								   timerInstanceP->deployedTimerP->intervalSecondsElement);
				messageNumber = MJB_SYNTAX_ELEMENT_DATA_CONTAINS_NULLS;
			}
			else
			{
				/* some other error has occurred */
				sprintf(msgBuffer, "Error occurred when trying to extract data from syntax "
					               "element specified for the '%S' attribute in message.  "
								   "The element path is '%S'",
								   ATTR_INTERVAL_SECONDS_MB,
								   timerInstanceP->deployedTimerP->intervalSecondsElement);
				messageNumber = MJB_SYNTAX_ELEMENT_DATA_RETRIEVAL_FAILURE;
			}
		}
		else
		{
			/* check that retrieved value is numeric */
			if (!isNumeric(valueBuffer))
			{
				sprintf(msgBuffer, "The value extracted from the syntax element specified "
					               "for the '%S' attribute in message contains non-numeric "
								   "data.  The element path is '%S', data is '%s'.",
								   ATTR_INTERVAL_SECONDS_MB,
								   timerInstanceP->deployedTimerP->intervalSecondsElement,
								   valueBuffer);
				dataValueP = valueBuffer;
				messageNumber = MJB_SYNTAX_ELEMENT_DATA_NON_NUMERIC;
				issueLogMessage = TRUE;
			}
			else
			{
				/* convert value to integer */
				derivedIntervalSeconds = atoi(valueBuffer);

				/* store derived interval value in timer instance */
				timerInstanceP->intervalSeconds = derivedIntervalSeconds;
				derivedIntervalValue = TRUE;
			}
		}

		if (issueLogMessage)
		{
			TRACE_MID(TRACE_ERROR, ("%s", msgBuffer)); 
			/* issue the appropriate log message */
			cciLog(&rc, 
				   CCI_LOG_ERROR, 
				   __FILE__, 
				   __LINE__, 
				   functionName, 
				   C.PLUGIN_MESSAGE_CATALOGUE, 
				   messageNumber, 
				   msgBuffer,
#if !defined(WIN32)
				   brokerName,
#endif
				   ATTR_TIMER_NAME_MB,
 				   /* convert MessageData to multi byte string */
				   UCToStr(timerInstanceP->deployedTimerP->timerName,
						   mbString,
						   sizeof(mbString),
						   NULL),
					dataValueP,  // this may be null, in which case its the terminator
				   (char *)0);
			                     

			if (dataValueP)
			{
				free(dataValueP);
			}
			free(timerInstanceP);

			TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
								   "terminal", message));
			/* 
			 * Propagate message to the 'failure' terminal if element doesn't exist or 
			 * has no data.
			 */
			retCode = propagateMessage(contextP,
									   C.TERM_FAILURE,
									   message,
									   destinationList,
									   exceptionList,
									   FALSE); /* this is not a new message */

			if (retCode != TT_SUCCESS)
			{
				TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
				RETHROW_IF_EXCEPTION(retCode);
				TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
			}

			TRACE_RETURN(RSN_013, TT_FAILED);
		}

	}
	else
	{
		timerInstanceP->intervalSeconds = timerInstanceP->deployedTimerP->intervalSeconds;
	}

	if (!timerInstanceP->intervalSeconds)
	{
		timerInstanceP->intervalSeconds = INFINITE;
	}

	if (timerInstanceP->intervalSeconds == INFINITE)
	{
		TRACE_MID(TRACE_INFO, ("%s%s value set to INFINITE (no timeout value)",
							   derivedIntervalValue ? "Derived " : "Literal ",
							   ATTR_INTERVAL_SECONDS_MB)); 
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("%s%s value set to %d seconds",
							   derivedIntervalValue ? "Derived " : "Literal ",
							   ATTR_INTERVAL_SECONDS_MB,
							   timerInstanceP->intervalSeconds)); 
	}

	/* if user specified a MessageData attribute, get required data */
	savedDataBufferLength = 0;
	savedDataBuffer = NULL;
	if (timerInstanceP->deployedTimerP->messageData)
	{
		/* if user specified a syntax tree element from which to copy data, locate it */
		if (timerInstanceP->deployedTimerP->retrieveMessageDataFromTree)
		{
			/* locate the specified element in the syntax tree and copy the data */
			copyFromElement = locateElementInMessage(message,
											timerInstanceP->deployedTimerP->messageData,
											FALSE,   /* don't create element */
											NULL,
											NULL);
			/* 
			 * Propagate message to the 'dataNotFound' terminal if element doesn't exist or 
			 * has no data.
			 */
			if (!copyFromElement)
			{
				sprintf(msgBuffer, "Failed to locate specified syntax element '%S' in message",
									timerInstanceP->deployedTimerP->messageData);

				cciLog(&rc, 
					   CCI_LOG_ERROR, 
					   __FILE__, 
					   __LINE__, 
					   functionName, 
					   C.PLUGIN_MESSAGE_CATALOGUE, 
					   MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT, 
					   msgBuffer,
#if !defined(WIN32)
					   brokerName,
#endif
 					   /* convert MessageData to multi byte string */
					   UCToStr(timerInstanceP->deployedTimerP->messageData,
							   mbString,
							   sizeof(mbString),
							   NULL),
					   (char *)0);

				free(timerInstanceP);

				TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
									   "terminal", message));
				retCode = propagateMessage(contextP,
										   C.TERM_FAILURE,
										   message,
										   destinationList,
										   exceptionList,
										   FALSE); /* this is not a new message */

				if (retCode != TT_SUCCESS)
				{
					TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
					RETHROW_IF_EXCEPTION(retCode);
					TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
				}

				TRACE_RETURN(RSN_013, TT_SUCCESS);
			}

			/* make a copy of the data (and elements) in/under this element */
			retCode = serialiseSyntaxElements(copyFromElement, 
				                              TRUE,  // store as string data 
				                              FALSE,  // don't ignore DTD information 
				                              &savedDataBuffer, 
				                              &savedDataBufferLength);
			if (retCode != TT_SUCCESS)
			{
				/* deal with common errors */
				if (retCode == TT_ELEMENT_HAS_NO_VALUE)
				{
					sprintf(msgBuffer, "Failed to locate any data in syntax element specified "
									   "for the '%S' attribute in message.  The element "
									   "path is '%S'",
									   ATTR_MESSAGE_DATA_MB,
									   timerInstanceP->deployedTimerP->messageData);
					messageNumber = MJB_SYNTAX_ELEMENT_CONTAINS_NO_DATA;
				}
				else
				if (retCode == TT_ELEMENT_DATA_CONTAINS_NULL_DATA_CHARACTER)
				{
					sprintf(msgBuffer, "Failed to extract any data from syntax element specified "
									   "for the '%S' attribute in message because the data "
									   "contains one or more null characters.  The element path "
									   "is '%S'",
									   ATTR_MESSAGE_DATA_MB,
									   timerInstanceP->deployedTimerP->messageData);
					messageNumber = MJB_SYNTAX_ELEMENT_DATA_CONTAINS_NULLS;
				}
				else
				{
					/* some other error has occurred */
					sprintf(msgBuffer, "Error occurred when trying to extract data from syntax "
									   "element specified for the '%S' attribute in message.  "
									   "The element path is '%S'",
									   ATTR_MESSAGE_DATA_MB,
									   timerInstanceP->deployedTimerP->messageData);
					messageNumber = MJB_SYNTAX_ELEMENT_DATA_RETRIEVAL_FAILURE;
				}

				/* write log message */
				cciLog(&rc, 
					   CCI_LOG_ERROR, 
					   __FILE__, 
					   __LINE__, 
					   functionName, 
					   C.PLUGIN_MESSAGE_CATALOGUE, 
					   messageNumber, 
					   msgBuffer,
#if !defined(WIN32)
					   brokerName,
#endif
					   ATTR_MESSAGE_DATA_MB,
 					   /* convert messageData to multi byte string */
					   UCToStr(timerInstanceP->deployedTimerP->messageData,
							   mbString,
							   sizeof(mbString),
							   NULL),
					   (char *)0);

				free(timerInstanceP);
				
				/* propagate to failure terminal if any errors */
				TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
									   "terminal", message));
				retCode = propagateMessage(contextP,
										   C.TERM_FAILURE,
										   message,
										   destinationList,
										   exceptionList,
										   FALSE); /* this is not a new message */

				if (retCode != TT_SUCCESS)
				{
					TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
					RETHROW_IF_EXCEPTION(retCode);
					TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
				}

				TRACE_RETURN(RSN_013, TT_SUCCESS);
			}
		}
		else
		{
			TRACE_MID(TRACE_INFO, ("Copying literal value '%S', length %d", 
									timerInstanceP->deployedTimerP->messageData,
									UClen(timerInstanceP->deployedTimerP->messageData)));
			savedDataBufferLength = UClen(timerInstanceP->deployedTimerP->messageData) + 1;
			savedDataBuffer = (char *)malloc(savedDataBufferLength);
			UCToStr(timerInstanceP->deployedTimerP->messageData,
					savedDataBuffer,
				    savedDataBufferLength,
					NULL);

			/* reset buffer length to length of string (w/o null) */
            savedDataBufferLength = strlen(savedDataBuffer);
		}

	}
	else
	{
		TRACE_MID(TRACE_INFO, ("No MessageData to insert")); 
	}

	/* update the timer with the serialised data */
	timerInstanceP->savedDataLength = savedDataBufferLength;
	if (scope==SCOPE_EXECUTION_GROUP)
	{
		timerInstanceP->savedDataP = savedDataBuffer;
		timerInstanceP->global = FALSE;
	}
	else
	{
		/* 
		 * Allocate a contiguous block of global storage to hold (in order) 
		 * - the global timer instance
		 * - the name of the Execution Group
		 * - the name of the Message Flow
		 * - the serialised data
		 * - the queue name
		 * - the timer name
		 */
		TRACE_MID(TRACE_INFO, ("sizeof(globalTimerInstance_t) = %d",
			                  sizeof(globalTimerInstance_t)));
		allocSize = sizeof(globalTimerInstance_t) +
			          strlen(executionGroupName) + 1 +
					  (UClen(contextP->messageFlowName) + 1) * sizeof(CciChar) +
                      timerInstanceP->savedDataLength +
					  (MQ_Q_NAME_LENGTH + 1) * sizeof(CciChar) +
			          (timerInstanceP->timerName ?
				      (strlen(timerInstanceP->timerName) + 1) : 0);

		TRACE_MID(TRACE_INFO, ("About to allocate %d bytes of global storage",
			                   allocSize));

		/* allocate global storage and copy existing buffer into this */
		globalTimerInstanceP = (globalTimerInstance_t *)shmMalloc(allocSize,
	 										                      FALSE,
																  scope==SCOPE_BROKER,
											                      &rc);
		CHECK_SHMEM_ERROR(rc, 
			              functionName, 
			              scope==SCOPE_BROKER, 
						  globalTimerInstanceP, 
						  globalTimerInstance_t *, 
						  allocSize, 
						  FALSE, /* lock is not held */
						  platformIsUnix,
						  TRUE);  // gc thread exists, so retry=TRUE 

		/* copy in the contents of the local timer */
		memcpy(globalTimerInstanceP, timerInstanceP, sizeof(timerInstance_t));

		/* append the additional data to global timer instance */
	    executionGroupNameP = (char *)globalTimerInstanceP + sizeof(globalTimerInstance_t);
		TRACE_MID(TRACE_INFO, ("Storage for executionGroupName located at %.8x",
			                   executionGroupNameP));

        messageFlowNameP = (CciChar *)(executionGroupNameP + strlen(executionGroupName) + 1);
		TRACE_MID(TRACE_INFO, ("Storage for messageFlowName located at %.8x",
			                   messageFlowNameP));

		dataP = (char *)messageFlowNameP + 
			               (UClen(contextP->messageFlowName) + 1) * sizeof(CciChar);
		if (timerInstanceP->savedDataLength)
		{
			TRACE_MID(TRACE_INFO, ("Storage for serialised data located at %.8x",
								   dataP));
		}

		queueNameP = (CciChar *)((char *)dataP + timerInstanceP->savedDataLength);
		TRACE_MID(TRACE_INFO, ("Storage for queueName located at %.8x",
			                   queueNameP));

        if (timerInstanceP->timerName)
		{
			timerNameP = (char *)((char *)queueNameP + 
					                 (MQ_Q_NAME_LENGTH + 1) * sizeof(CciChar));
			TRACE_MID(TRACE_INFO, ("Storage for timer name located at %.8x",
								   timerNameP));
		}
		else
		{
			timerNameP = NULL;
			TRACE_MID(TRACE_INFO, ("Timer does not have a name, no global storage required"));
		}

		strcpy(executionGroupNameP, executionGroupName);
		UCcpy(messageFlowNameP, contextP->messageFlowName);
		if (timerInstanceP->savedDataLength)
		{
			memcpy(dataP, savedDataBuffer, timerInstanceP->savedDataLength);
		}
		UCcpy(queueNameP, timerInstanceP->deployedTimerP->queueName);
        if (timerInstanceP->timerName)
		{
			strcpy(timerNameP, timerInstanceP->timerName);
		}

		/* since this is a global timer, we must use handles instead of pointers */
		globalTimerInstanceP->executionGroupNameP = SHM_ADDRESS2HANDLE(executionGroupNameP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		globalTimerInstanceP->messageFlowNameP = SHM_ADDRESS2HANDLE(messageFlowNameP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		globalTimerInstanceP->queueNameP = SHM_ADDRESS2HANDLE(queueNameP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		if (timerInstanceP->savedDataLength)
		{
			globalTimerInstanceP->savedDataP = SHM_ADDRESS2HANDLE(dataP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		globalTimerInstanceP->timerName = SHM_ADDRESS2HANDLE(timerNameP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		globalTimerInstanceP->global = TRUE;
		globalTimerInstanceP->executionGroupPID = currentPID;
		globalTimerInstanceP->instanceLength = allocSize;
		globalTimerInstanceP->retrieveMessageDataFromTree =
			                      timerInstanceP->deployedTimerP->retrieveMessageDataFromTree;

		free(savedDataBuffer);
		if (timerInstanceP->timerName)
		{
            free(timerInstanceP->timerName);
		}
		free(timerInstanceP);  /* free timer in local memory */

		/* use same pointer to address either local or global for simplicity */
		timerInstanceP = (timerInstance_t *)globalTimerInstanceP;
	}

	if (timerInstanceP->savedDataLength)
	{
		TRACE_MID(TRACE_INFO, ("Timer updated with serialised data at %.8x",
								scope==SCOPE_EXECUTION_GROUP ? timerInstanceP->savedDataP : 
								SHM_HANDLE2ADDRESS(globalTimerInstanceP->savedDataP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		TRACE_MIDHEX(TRACE_INFO, scope==SCOPE_EXECUTION_GROUP ? timerInstanceP->savedDataP : 
								 SHM_HANDLE2ADDRESS(globalTimerInstanceP->savedDataP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform),
									 scope==SCOPE_EXECUTION_GROUP ? timerInstanceP->savedDataLength :
									 globalTimerInstanceP->savedDataLength);
	}

	TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
	threadMutexLock(timerChainMutex);

	/* add the timer to the deployed create node */
	instanceAnchorP = &deployedTimerP->instanceAnchor;
	INSERT_AFTER_IN_LINKED_LIST(instanceAnchorP, timerInstanceP);

	/* set the timer timeout and add to alarm chain */
	if (timerInstanceP->intervalSeconds != INFINITE)
	{
	    setTimerTimeoutAlarm(timerInstanceP);
	
		/* signal the alarm thread to process new timer instance */
		threadCondSignal(alarmThreadWorkCondVar);
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Timer has no timeout value"));
	}

	/* add the timer to required indexes and chains */
	updateAllIndexes(timerInstanceP, 
		             scope,
					 TRUE);  // remove any matching timers

	TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
	threadMutexUnlock(timerChainMutex);

    /* propagate the message to the output terminal */
	TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'out' "
			               "terminal", message));
	retCode = propagateMessage(contextP,
							   C.TERM_OUT,
							   message,
							   destinationList,
							   exceptionList,
							   FALSE); /* this is not a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		RETHROW_IF_EXCEPTION(retCode);
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	TRACE_RETURN(RSN_013, TT_SUCCESS);
}

/******************************************************************************/
/* updateAllIndexes                                                           */
/******************************************************************************/
void updateAllIndexes(IN timerInstance_t *newInstanceP,
					  IN scope_t scope,
					  IN bool_t replaceMatchingInstances)
{
	timerInstance_t *currentInstanceP;
	char *timerName;
	globalTimerInstance_t *currentGlobalInstanceP,
                          *newGlobalInstanceP,
						  *nextGlobalInstanceP;
	int retCode,
		numMatches;
	bool_t nullCorrelId,
		   nullMsgId;
	char nullId[MQ_MSG_ID_LENGTH];

    TRACE_ENTRY("updateAllIndexes", ("newInstanceP=%.8x, scope=%s, "
		                             "replaceMatchingInstances=%s", 
		                             newInstanceP,
									 scope==SCOPE_EXECUTION_GROUP ? "SCOPE_EXECUTION_GROUP" :
									 scope==SCOPE_BROKER ? "SCOPE_BROKER" : "SCOPE_PLATFORM",
									 replaceMatchingInstances ? "TRUE" : "FALSE"));


	/* Check for null keys, i.e. either MsgId or CorrelId set to binary zeroes */
	nullCorrelId = FALSE;
	nullMsgId = FALSE;
	memset(nullId, 0, sizeof(nullId));
	if (!memcmp(newInstanceP->correlId, nullId, sizeof(nullId)))
	{
		nullCorrelId = TRUE;
        TRACE_MID(TRACE_INFO, ("CorrelId is null"));
	}
	if (!memcmp(newInstanceP->msgId, nullId, sizeof(nullId)))
	{
		nullMsgId = TRUE;
        TRACE_MID(TRACE_INFO, ("MsgId is null"));
	}

	/* 
	 * Assume caller has timerChainMutex locked on entry... 
	 *
	 * NB: Even though some of the timers on the local chain may in fact reside in shared
	 * storage, we can read these without taking any locks.  This is because they are only
	 * ever deleted by the owning process (and that's us).  Furthermore, we can also 
	 * update the local chain pointers of these timers without a lock since these fields
	 * are only ever referenced by the owning process (again, thats us).
	 */

	/* 
	 * We always update the local chains since the Execution Group will always be in scope.
	 */

	/* if timer has a name, then this is the only chain that we need to belong to */
	if (!newInstanceP->global && newInstanceP->timerName || 
	     newInstanceP->global && (shmemHandle_t)newInstanceP->timerName != SHM_NULL_HANDLE)
	{
		if (replaceMatchingInstances)
		{
			/* remove any matching names in the local index */
			numMatches = 0;
			if (newInstanceP->global)
			{
				timerName = 
					SHM_HANDLE2ADDRESS(
					    ((globalTimerInstance_t *)newInstanceP)->timerName,
						char *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			}
			else
			{
				timerName = newInstanceP->timerName;
			}
			retCode = searchIndexAndCancel(timerName,
				                           INDEX_NAME,
				                           scope,
										   &numMatches);
			if (numMatches)
			{
				TRACE_MID(TRACE_INFO, ("Existing entry with matching name removed from "
					                   "local%s index%s",
									   newInstanceP->global ? " and global" : "",
									   newInstanceP->global ? "es" : ""));
			}
			else
			{
				TRACE_MID(TRACE_INFO, ("No entries to replace"));
			}
		}

		/* 
		 * Update the timer name index
		 */
		currentInstanceP = nameIndexAnchorP;

		/* insert timer instance into chain in name order */
		while(currentInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentInstanceP));
			if (!currentInstanceP->nextNameP ||
				strcmp(newInstanceP->timerName, 
					   currentInstanceP->nextNameP->timerName) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in name index"));
				break;
			}
			currentInstanceP = currentInstanceP->nextNameP;
		}
		INSERT_AFTER_IN_NAME_CHAIN(currentInstanceP, newInstanceP);
	}
	else
	{
		if (replaceMatchingInstances)
		{
			// remove any entries with matching MsgId AND CorrelId
			TRACE_MID(TRACE_INFO, ("Using MsgId+CorrelId value to search Both index"));
			retCode = searchIndexAndCancel(newInstanceP->msgId,
				                           INDEX_BOTH,
				                           scope,
										   &numMatches);
			
			if (numMatches)
			{
				TRACE_MID(TRACE_INFO, ("Existing entry with matching name removed from "
					                   "local%s index%s",
									   newInstanceP->global ? " and global" : "",
									   newInstanceP->global ? "es" : ""));
			}
			else
			{
				TRACE_MID(TRACE_INFO, ("No entries to replace"));
			}
		}

		/* 
		 * Update the MsgId index
		 */
		currentInstanceP = msgIdIndexAnchorP;

		/* insert timer instance into chain in message id order */
		while(currentInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentInstanceP));
			if (!currentInstanceP->nextMsgIdP ||
				memcmp(newInstanceP->msgId, 
					   currentInstanceP->nextMsgIdP->msgId, MQ_MSG_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in MsgId index"));
				break;
			}
			currentInstanceP = currentInstanceP->nextMsgIdP;
		}
		INSERT_AFTER_IN_MSG_ID_CHAIN(currentInstanceP, newInstanceP);

		/* 
		 * Update the CorrelId index
		 */
		currentInstanceP = correlIdIndexAnchorP;

		/* insert timer instance into chain in correl id order */
		while(currentInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentInstanceP));
			if (!currentInstanceP->nextCorrelIdP ||
				memcmp(newInstanceP->correlId, 
					   currentInstanceP->nextCorrelIdP->msgId, MQ_CORREL_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in CorrelId index"));
				break;
			}
			currentInstanceP = currentInstanceP->nextCorrelIdP;
		}
		INSERT_AFTER_IN_CORREL_ID_CHAIN(currentInstanceP, newInstanceP);

		/* 
		 * Update the Both index
		 */
		currentInstanceP = bothIndexAnchorP;

		/* insert timer instance into chain in Msgid+CorrelId order */
		/* Since MsgId and CorrelId are contiguous in storage, we can simply compare 48 bytes */
		while(currentInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentInstanceP));
			if (!currentInstanceP->nextBothP ||
				memcmp(newInstanceP->msgId, 
					   currentInstanceP->nextBothP->msgId, MQ_MSG_ID_LENGTH+MQ_CORREL_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in MsgId+CorrelId index"));
				break;
			}
			currentInstanceP = currentInstanceP->nextBothP;
		}
		INSERT_AFTER_IN_BOTH_CHAIN(currentInstanceP, newInstanceP);
	}

	/**********************************************************************************/
	/**********************************************************************************/
	/* Update global chains                                                           */
	/**********************************************************************************/
	/**********************************************************************************/
	if (scope==SCOPE_EXECUTION_GROUP)
	{
		TRACE_RETURN_VOID(RSN_001);
	}

	TRACE_MID(TRACE_INFO, ("Updating global indexes"));

    newGlobalInstanceP = (globalTimerInstance_t *)newInstanceP;

	/* lock shared memory */
	shmLock(scope==SCOPE_BROKER);

	/* Since we are chaining global storage, we must use handles as opposed to pointers */
	newGlobalInstanceP->nextGlobalP = SHM_NULL_HANDLE;
	newGlobalInstanceP->prevGlobalP = SHM_NULL_HANDLE;
	newGlobalInstanceP->nextGlobalMsgIdP = SHM_NULL_HANDLE;
	newGlobalInstanceP->prevGlobalMsgIdP = SHM_NULL_HANDLE;
	newGlobalInstanceP->nextGlobalCorrelIdP = SHM_NULL_HANDLE;
	newGlobalInstanceP->prevGlobalCorrelIdP = SHM_NULL_HANDLE;
	newGlobalInstanceP->nextGlobalBothP = SHM_NULL_HANDLE;
	newGlobalInstanceP->prevGlobalBothP = SHM_NULL_HANDLE;
	newGlobalInstanceP->nextGlobalNameP = SHM_NULL_HANDLE;
	newGlobalInstanceP->prevGlobalNameP = SHM_NULL_HANDLE;


	/* if timer has a name, then this is the only chain that we need to belong to */
	if (newGlobalInstanceP->timerName != SHM_NULL_HANDLE)
	{
		/* 
		 * Update the global name index
		 */
		currentGlobalInstanceP = &(scope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor;

		/* insert timer instance into chain in name order */
		while(currentGlobalInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentGlobalInstanceP));
			if (currentGlobalInstanceP->nextGlobalNameP == SHM_NULL_HANDLE)
			{
				break;
			}

			/* address next instance in chain */
			nextGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalNameP,
					               globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			if (strcmp(SHM_HANDLE2ADDRESS(currentGlobalInstanceP->timerName, char *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform),
				       SHM_HANDLE2ADDRESS(nextGlobalInstanceP->timerName, char *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in global MsgId index"));
				break;
			}
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalNameP, globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		INSERT_AFTER_IN_GLOBAL_TIMER_NAME_CHAIN(currentGlobalInstanceP, newGlobalInstanceP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

	}
	else
	{
		/* 
		 * Update the global MsgId index
		 */
		currentGlobalInstanceP = &(scope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor;

		/* insert timer instance into chain in message id order */
		while(currentGlobalInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentGlobalInstanceP));
			if (!SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform) ||
				memcmp(newGlobalInstanceP->msgId, 
					   SHM_HANDLE2ADDRESS(
						  currentGlobalInstanceP->nextGlobalMsgIdP, globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)->msgId, 
						  MQ_MSG_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in global MsgId index"));
				break;
			}
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		INSERT_AFTER_IN_GLOBAL_TIMER_MSG_ID_CHAIN(currentGlobalInstanceP, newGlobalInstanceP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		/* 
		 * Update the CorrelId index
		 */
		currentGlobalInstanceP = &(scope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor;

		/* insert timer instance into chain in correl id order */
		while(currentGlobalInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentGlobalInstanceP));
			if (!SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform) ||
				memcmp(newGlobalInstanceP->correlId, 
					   SHM_HANDLE2ADDRESS(
						  currentGlobalInstanceP->nextGlobalCorrelIdP, globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)->correlId, 
						  MQ_MSG_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in global CorrelId index"));
				break;
			}
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		INSERT_AFTER_IN_GLOBAL_TIMER_CORREL_ID_CHAIN(currentGlobalInstanceP, newGlobalInstanceP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		/* 
		 * Update the Both index
		 */
		currentGlobalInstanceP = &(scope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor;

		/* insert timer instance into chain in Msgid+CorrelId order */
		/* Since MsgId and CorrelId are contiguous in storage, we can simply compare 48 bytes */
		while(currentGlobalInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentGlobalInstanceP));

			if (!SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform) ||
				memcmp(newGlobalInstanceP->msgId, 
					   SHM_HANDLE2ADDRESS(
						  currentGlobalInstanceP->nextGlobalBothP, globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)->msgId, 
						  MQ_MSG_ID_LENGTH+MQ_CORREL_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in global MsgId+CorrelId index"));
				break;
			}
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, globalTimerInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		INSERT_AFTER_IN_GLOBAL_TIMER_BOTH_CHAIN(currentGlobalInstanceP, newGlobalInstanceP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	}

	/* unlock shared memory */
	shmUnlock(scope==SCOPE_BROKER);

	TRACE_RETURN_VOID(RSN_001);
}

/* local functions */
/*
 * isNumeric
 */
bool_t isNumeric(char *string)
{
	unsigned int i;

	// check that each character is an integer value 
	for (i=0; i<strlen(string); i++)
	{
		if (!strchr("0123456789", *(string+i)))
		{
			return(FALSE);
		}
	}
	return(TRUE);
}

