/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : garbage.c                                                */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Provides global storage garbage collection functions for */
/*                   use by the timer node.                                   */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <timer.h> /* Specific header for this dll                            */

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <threads.h>
#include <time.h>

#define DEFAULT_SLEEP_SECONDS 20
/******************************************************************************/
/* sharedMemoryGarbageCollection                                              */
/******************************************************************************/
threadFunction_t sharedMemoryGarbageCollection(void *dataP)
{
	time_t currentTime;
	int timeToWait,
		numTimersFound,
		rc,
		brokerPercentFree = 100,
		platformPercentFree = 100,
		minPercentFree;
	double fPercentFree;
	globalTimerInstance_t *deadAnchorBrokerP = NULL,
		                  *deadAnchorPlatformP = NULL,
		                  *currentInstanceP,
						  *nextInstanceP;
    static char *functionName="sharedMemoryGarbageCollection";

    TRACE_ENTRY("sharedMemoryGarbageCollection", ("dataP=%.8x", dataP));

	/* let the parent thread know that we're up and running */
	shmemThreadInitialised = TRUE;

	/* lock the alarm thread sync mutex and post the parent thread */
    TRACE_MID(TRACE_INFO, ("About to signal parent for initialisation complete"));
	threadMutexLock(shmemThreadSyncMutex);
	threadCondSignal(shmemThreadSyncCondVar);
	threadMutexUnlock(shmemThreadSyncMutex);

	/* address the dead chains */
	if (!deadAnchorBrokerP && sharedMemoryCBP_Broker)
	{
		deadAnchorBrokerP = &sharedMemoryCBP_Broker->globalDeadInstanceAnchor;
	}
	if (!deadAnchorPlatformP && sharedMemoryCBP_Platform)
	{
		deadAnchorPlatformP = &sharedMemoryCBP_Platform->globalDeadInstanceAnchor;
	}

	/* 
	 * Periodically check the global storage dead chain for any postits to be removed
	 */
	threadMutexLock(shmemThreadWorkMutex);

	/* keep processing global storage dead chain until requested to shutdown */
	while(!shutdownShmemRequested)
	{
		/* address the dead chains */
		if (sharedMemoryCBP_Broker)
		{
			deadAnchorBrokerP = &sharedMemoryCBP_Broker->globalDeadInstanceAnchor;
		}
		if (sharedMemoryCBP_Platform)
		{
			deadAnchorPlatformP = &sharedMemoryCBP_Platform->globalDeadInstanceAnchor;
		}

		/* take a peek at the Broker dead chain (dirty read) to avoid locking shared memory */
		numTimersFound = 0;
		if (deadAnchorBrokerP) {
			// get the current heap stats 
			brokerPercentFree = shmGetPercentFree(TRUE,  // PRETEND we've got the lock
			                                      TRUE); // scope is broker 
		    if (deadAnchorBrokerP->nextGlobalP != SHM_NULL_HANDLE)
			{
			    /* 
			     * Ok, looks like theres something on the broker chain, so lock global storage 
			     * and check again.
			     */
			    /* 
			     * The timer chain and the shared memory dead chain are related in that if we find
			     * anything to release on the dead chain, we also need to remove it from the
			     * local chains, so we always take the timerChainMutex first to avoid deadlock.
			     */
			    TRACE_MID(TRACE_INFO, ("Processing Broker dead chain"));
			    TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
			    threadMutexLock(timerChainMutex);

			    TRACE_MID(TRACE_INFO, ("About to take Broker global lock"));
			    processLockLock(shmemLock_Broker);

			    // get the actual current heap stats before we start freeing storage
			    brokerPercentFree = shmGetPercentFree(TRUE,  // we've got the lock
			                                          TRUE); // scope is broker 

			    currentInstanceP = SHM_HANDLE2ADDRESS(deadAnchorBrokerP->nextGlobalP, 
				                                      globalTimerInstance_t *,
							  					      sharedMemorySegment_Broker);
			    while(currentInstanceP)
				{
				    nextInstanceP = SHM_HANDLE2ADDRESS(currentInstanceP->nextGlobalP,
					                                   globalTimerInstance_t *,
												       sharedMemorySegment_Broker);

                    /* skip this timer if it doesn't belong to our execution group */
				    if (currentInstanceP->executionGroupPID != currentPID)
					{
					    TRACE_MID(TRACE_INFO, ("Timer owned by process %d - timer ignored", 
						                        currentInstanceP->executionGroupPID));
				        currentInstanceP = nextInstanceP;
					    continue;
					}

				    /* process one of 'our' timers */
				    numTimersFound++;
				    TRACE_MID(TRACE_INFO, ("We own this timer, so unchain and free it"));
				    REMOVE_FROM_GLOBAL_TIMER_CHAIN(currentInstanceP, sharedMemorySegment_Broker);

				    /* remove from local chains too... */
                    unchainTimer((timerInstance_t *)currentInstanceP);

				    /* finally, free up the shared memory */
			        rc = shmFree(currentInstanceP, 
					             TRUE, /* we've already got the shmem lock */
						         TRUE); // broker scope 

		            CHECK_SHMFREE_ERROR(rc, 
			                            functionName, 
						                platformIsUnix);  

				    currentInstanceP = nextInstanceP;
				}

			    TRACE_MID(TRACE_INFO, ("About to release Broker global lock"));
			    processLockUnlock(shmemLock_Broker);

			    TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
			    threadMutexUnlock(timerChainMutex);
			}
		}

		/* take a peek at the Platform dead chain (dirty read) to avoid locking shared memory */
		numTimersFound = 0;
		if (deadAnchorPlatformP) {
			// get the current heap stats 
			brokerPercentFree = shmGetPercentFree(TRUE,  // PRETEND we've got the lock
			                                      FALSE); // scope is platform 
			if (deadAnchorPlatformP->nextGlobalP != SHM_NULL_HANDLE)
			{
				/* 
				 * Ok, looks like theres something on the platform chain, so lock global storage 
				 * and check again.
				 */
				/* 
				 * The timer chain and the shared memory dead chain are related in that if we find
				 * anything to release on the dead chain, we also need to remove it from the
				 * local chains, so we always take the timerChainMutex first to avoid deadlock.
				 */
				TRACE_MID(TRACE_INFO, ("Processing Platform dead chain"));
				TRACE_MID(TRACE_INFO, ("About to lock timerChainMutex"));
				threadMutexLock(timerChainMutex);

				TRACE_MID(TRACE_INFO, ("About to take Platform global lock"));
				processLockLock(shmemLock_Platform);

				// get the actual current heap stats before we start freeing storage
				brokerPercentFree = shmGetPercentFree(TRUE,  // we've got the lock
													  FALSE); // scope is broker 

				currentInstanceP = SHM_HANDLE2ADDRESS(deadAnchorPlatformP->nextGlobalP, 
													  globalTimerInstance_t *,
													  sharedMemorySegment_Platform);
				while(currentInstanceP)
				{
					nextInstanceP = SHM_HANDLE2ADDRESS(currentInstanceP->nextGlobalP,
													   globalTimerInstance_t *,
													   sharedMemorySegment_Platform);

					/* skip this timer if it doesn't belong to our execution group */
					if (currentInstanceP->executionGroupPID != currentPID)
					{
						TRACE_MID(TRACE_INFO, ("Timer owned by process %d - timer ignored", 
												currentInstanceP->executionGroupPID));
						currentInstanceP = nextInstanceP;
						continue;
					}

					/* process one of 'our' timers */
					numTimersFound++;
					TRACE_MID(TRACE_INFO, ("We own this timer, so unchain and free it"));
					REMOVE_FROM_GLOBAL_TIMER_CHAIN(currentInstanceP, sharedMemorySegment_Platform);

					/* remove from local chains too... */
					unchainTimer((timerInstance_t *)currentInstanceP);

					/* finally, free up the shared memory */
					rc = shmFree(currentInstanceP, 
								 TRUE, /* we've already got the shmem lock */
								 FALSE); // platform scope 

					CHECK_SHMFREE_ERROR(rc, 
										functionName, 
										platformIsUnix);  

					currentInstanceP = nextInstanceP;
				}

				TRACE_MID(TRACE_INFO, ("About to release Broker global lock"));
				processLockUnlock(shmemLock_Broker);

				TRACE_MID(TRACE_INFO, ("About to unlock timerChainMutex"));
				threadMutexUnlock(timerChainMutex);
			}
		}

		/* go to sleep for while... */
		/* reset the garbage thread to pop at next expiry interval */
		time(&currentTime);
        TRACE_MID(TRACE_INFO, ("currentTime=%d", currentTime));

		// calculate time to wait - this is directly linked to the amount of free storage left.
		// The more free storage remaining, the longer we sleep and vice versa.
		minPercentFree = platformPercentFree < brokerPercentFree ? platformPercentFree : brokerPercentFree; 

		fPercentFree = (double)(minPercentFree * DEFAULT_SLEEP_SECONDS) / (double)100;

        //TRACE_MID(TRACE_INFO, ("fPercentFree=%f", fPercentFree));

		timeToWait = (int)(fPercentFree);
		// round up the result
		if (fPercentFree - timeToWait >= 0.5) timeToWait++;

		if (timeToWait == 0) {
			timeToWait = 1;  // make sure we wait at least one second
		}
        TRACE_MID(TRACE_INFO, ("Time to wait set to %d seconds", timeToWait));

        TRACE_MID(TRACE_INFO, ("Next expiry interval time is %d", currentTime+timeToWait));

		// convert to milliseconds
		timeToWait *= 1000;

	    /* wait for next alarm to pop or a new element to appear on the chain */
        TRACE_MID(TRACE_INFO, ("Garbage collection thread about to sleep for expiry interval"));
		sharedMemoryGarbageCollectorHasRun = TRUE;
		rc = threadCondTimedWait(shmemThreadWorkCondVar, timeToWait);
		if (rc == THREAD_TIMED_OUT) {
            TRACE_MID(TRACE_INFO, ("Garbage collection thread woken normally"));
		}
		else {
            TRACE_MID(TRACE_INFO, ("Garbage collection thread woken prematurely"));
		}
	}

    TRACE_MID(TRACE_INFO, ("Shared memory garbage collection thread terminating"));

	threadMutexUnlock(shmemThreadWorkMutex);

    TRACE_RETURN_TYPE(RSN_001, FUNCTION_RETURN_TYPE, 0);
}
